#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <iomanip.h>
#include <time.h>
#include "parser.hpp"
#include "filo.hpp"

const char ConfigFileName[] = "filo.ctl";
const int VerMajor = 1, // major version number
          VerMinor = 1; // major version number
          // version == VerMajor.VerMinor
          
  int  QuietMode;
  char AppDir[MAX_PATH];
  char Address[MAX_LINE];        
  char NewFileName[MAX_LINE];    
  char ResultFileName[MAX_LINE]; 
  char AreasFileName[MAX_LINE];
  int  DeleteFile;               
  int  UseShortOrigin;           
  int  UseShortUplink;
  int  AddEmptyLine;
  char PathTemplate[MAX_LINE];
  int  SkipDescChars;
  char SkipSet[MAX_LINE];
  int  ReplaceSoftCR;
  int  CreateFileFlags;
  char FileFlagsDir[MAX_PATH];
  char HeaderFileName[MAX_PATH];
  char TailFileName[MAX_PATH];

  int             TplsNumber;
  TemplateStruct *PathTpl;

  void main( int argc, char **argv ) {
    int i;

    //  ⠫ ᪠
    for (i = strlen(argv[0])-1; (i >= 0) && (argv[0][i] != '\\'); i--)
      ;
    if (i >= 0) { // 諨 
      memcpy( AppDir, argv[0], i+1 );
      AppDir[i+1] = '\0';
    }
    else
      AppDir[0] = '\0'; // ⠫ । 

    QuietMode = 0;
    //  㪠 ⢥ : -q (quiet mode/娩 ०)
    if (argc > 1) {
      for (i = 1; i < argc; i++)
        if (stricmp( argv[i], "-q" ) == 0)
          break;
      // ?
      if (i < argc)
        QuietMode = 1;
    }

    PrintLogo();         // 뢮 ଠ  ணࠬ

    DefaultControls();  // ⠭ ࠬ  㬮砭
    ReadControls();     // 뢠 ࠬ
    PrepareTemplates(); // ⮢ 蠡
    if (SkipDescChars)
      PrepareSkipTable(); // ⮢ ⠡ ᨬ  ய᪠
    MakeReport();       // ᮧ 
  }

  void DefaultControls() {
    strcpy( Address, "-1:-1/-1");
    strcpy( NewFileName, "newfile.txt" );
    strcpy( ResultFileName, "result.txt" );
    strcpy( AreasFileName, "areas.ctl" );
    DeleteFile = 0;
    UseShortOrigin = 0;
    UseShortUplink = 0;
    AddEmptyLine   = 0;
    *PathTemplate = '\0';
    SkipDescChars = 0;
    strcpy( SkipSet, "0-31,''-''" );
    ReplaceSoftCR = 0;
    CreateFileFlags = 0;
    strcpy( FileFlagsDir, "" );
    *HeaderFileName = '\0';
    *TailFileName   = '\0';
  }

  void ReadControls() {
    char      CtlName[MAX_PATH];
    ifstream  in;
    char      buf[MAX_LINE];

    strcpy( CtlName, AppDir );
    strcat( CtlName, ConfigFileName );

    in.open( CtlName );
    if (in) {
      if (QuietMode == 0)
        cout << "Reading control file " << CtlName << "...";
      while (!in.eof()) {
          in.getline( buf, MAX_LINE );
          if (!in)
              break;
          if (ParseString( buf, "Address", Address ))
              continue;
          if (ParseString( buf, "NewFileName", NewFileName ))
              continue;
          if (ParseString( buf, "ResultFileName", ResultFileName ))
              continue;
          if (ParseString( buf, "AreasFileName", AreasFileName ))
              continue;
          if (ParseYesNo( buf, "DeleteFile", DeleteFile ))
              continue;
          if (ParseYesNo( buf, "UseShortOrigin", UseShortOrigin ))
              continue;
          if (ParseYesNo( buf, "UseShortUplink", UseShortUplink ))
              continue;
          if (ParseYesNo( buf, "AddEmptyLine", AddEmptyLine ))
              continue;
          if (ParseQuotedString( buf, "PathTemplate", PathTemplate ))
              continue;
          if (ParseYesNo( buf, "SkipDescChars", SkipDescChars ))
              continue;
          if (ParseString( buf, "SkipSet", SkipSet ))
              continue;
          if (ParseYesNo( buf, "ReplaceSoftCR", ReplaceSoftCR ))
              continue;
          if (ParseYesNo( buf, "CreateFileFlags", CreateFileFlags ))
              continue;
          if (ParseString( buf, "FileFlagsDir", FileFlagsDir ))
              continue;
          if (ParseString( buf, "HeaderFileName", HeaderFileName ))
              continue;
          if (ParseString( buf, "TailFileName", TailFileName ))
              continue;
      }
      in.close();
      if (QuietMode == 0)
        cout << "Done.\n";
    }
    else {
      if (QuietMode == 0)
        cout << "[!] Failed reading contol file " << CtlName << endl
             << "[!] Using default settings.\n";
    }
  }

  static char SpaceLine[26] = "                         ";

  void MakeReport() {
    ifstream        in;
    ofstream        out;
    char            FName[MAX_PATH];
    char            buf[MAX_LINE];
    char            Area[MAX_LINE];
    char            File[MAX_LINE];
    char            Desc[MAX_LINE];
    char            cSize[MAX_LINE];
    int             Size;
    char            Origin[MAX_LINE];
    char            Uplink[MAX_LINE];
    AreaListStruct *AreaList, *LastArea, *CurArea;
    int             i, j;
    int             AreasNumber;
    AreaStruct     *Areas;
    FileListStruct *CurFile, *LastFile;
    int             TotalSize, AreaSize, TotalNumber;

    //   뢠 

    //    ?
    if (NewFileName[1] == ':')
      strcpy( FName, NewFileName );
    else {
      strcpy( FName, AppDir );
      strcat( FName, NewFileName );
    }

    in.open( FName );

    if (!in) {
      if (QuietMode == 0)
          cout << "[!] Failed reading input file " << FName << endl;
      return;
    }

    if (QuietMode == 0)
      cout << "Reading input file " << FName << "...";

    /*
        keyword sequence is
        Area
        File
        Size
        Desc
        Origin
        Uplink
    */
    AreaList = 0;
    AreasNumber = 0;
    while (!in.eof()) {

      // reading area name

        in.getline( buf, MAX_LINE );
        if (!in)
            break;

        if (!ParseString( buf, "Area", Area ))
            continue;
        
        // 饬  
  
        CurArea = AreaList;
        while (CurArea != 0) {
          if (strcmp(CurArea->Area.Name, Area) == 0)
            break;
          CurArea = CurArea->Next;
        }
        if (CurArea == 0) {
          if (AreaList == 0) {
            AreaList = new AreaListStruct;
            LastArea = AreaList;
          }
          else {
            LastArea->Next = new AreaListStruct;
            LastArea = LastArea->Next;
          }
          LastArea->Next = 0;
          LastArea->Area.Name = new char[strlen(Area) + 1];
          strcpy( LastArea->Area.Name, Area );
          LastArea->FileList = 0;
          LastArea->Area.FilesNumber = 0;
          AreasNumber++;
          CurArea  = LastArea;
        }
  
        // 뢠  䠩
  
        in.getline( buf, MAX_LINE );
        if (!ParseString( buf, "File", File ))
            continue;
  
        // 뢠 ࠧ

        in.getline( buf, MAX_LINE );
        if (!ParseString( buf, "Size", cSize ))
            continue;
        sscanf( cSize, "%d", &Size );
  
        // 뢠 ᠭ
  
        in.getline( buf, MAX_LINE );
        if (!ParseEmptyString( buf, "Desc", Desc ))
            continue;
        if (SkipDescChars)
          DoSkipChars(Desc);
        if (ReplaceSoftCR)
            DoReplaceSoftCR(Desc);
        SkipSpaces(Desc);
  
        // 뢠 Origin
  
        in.getline( buf, MAX_LINE );
        if (!ParseString( buf, "Origin", Origin ))
            continue;
  
        // 뢠 Uplink
  
        in.getline( buf, MAX_LINE );
        if (!ParseString( buf, "Uplink", Uplink ))
            continue;
  
        // 塞  䠩
  
        if (CurArea->FileList == 0) {
          CurArea->FileList = new FileListStruct;
          CurArea->LastFile = CurArea->FileList;
        }
        else {
          CurArea->LastFile->Next = new FileListStruct;
          CurArea->LastFile = CurArea->LastFile->Next;
        }
        CurArea->LastFile->Next = 0;
        CurArea->LastFile->File.Name = new char[strlen(File) + 1];
        strcpy( CurArea->LastFile->File.Name, File );
        CurArea->LastFile->File.Size = Size;
        CurArea->LastFile->File.Desc = new char[strlen(Desc) + 1];
        strcpy( CurArea->LastFile->File.Desc, Desc );
        if (UseShortOrigin)
          TruncateAddress( Origin );
        CurArea->LastFile->File.Origin = new char[strlen(Origin) + 1];
        strcpy( CurArea->LastFile->File.Origin, Origin );
        if (UseShortUplink)
          TruncateAddress( Uplink );
        CurArea->LastFile->File.Uplink = new char[strlen(Uplink) + 1];
        strcpy( CurArea->LastFile->File.Uplink, Uplink );
        CurArea->Area.FilesNumber++;
    }

    in.close();

    if (QuietMode == 0)
      cout << "Done.\n";

    // 塞 室 䠩

    if (DeleteFile) {
      if (NewFileName[1] == ':')
        strcpy( FName, NewFileName );
      else {
        strcpy( FName, AppDir );
        strcat( FName, NewFileName );
      }
      if (remove( FName ) == 0) { // 㤠筮 㤠
        if (QuietMode == 0) {
          cout << "File " << FName << " was successfully deleted.\n";
        }
      }
      else
        if (QuietMode == 0) {
          cout << "Failed deleting file " << FName << endl;
        }
    }

    // ८ࠧ㥬 ᯨ᪨  ᨢ

    if (AreasNumber == 0)
      return;

    Areas = new AreaStruct[AreasNumber];
    CurArea = AreaList;
    for (i = 0; i < AreasNumber; i++) {
      Areas[i].Name = CurArea->Area.Name;
      Areas[i].FilesNumber = CurArea->Area.FilesNumber;
      Areas[i].Desc = 0;
      if (Areas[i].FilesNumber > 0) {
        Areas[i].Files = new FileStruct[Areas[i].FilesNumber];
        CurFile = CurArea->FileList;
        for (j = 0; j < Areas[i].FilesNumber; j++) {
          Areas[i].Files[j].Name = CurFile->File.Name;
          Areas[i].Files[j].Size = CurFile->File.Size;
          Areas[i].Files[j].Desc = CurFile->File.Desc;
          Areas[i].Files[j].Origin = CurFile->File.Origin;
          Areas[i].Files[j].Uplink = CurFile->File.Uplink;
          LastFile = CurFile->Next;
          delete CurFile;
          CurFile = LastFile;
        }
        SortFiles( Areas[i].Files, Areas[i].FilesNumber );
      }
      LastArea = CurArea->Next;
      delete CurArea;
      CurArea = LastArea;
    }
    SortAreas( Areas, AreasNumber );

    // ⠥  ᠭ ⥩
    if (AreasFileName[1] == ':')
      strcpy( FName, AreasFileName );
    else {
      strcpy( FName, AppDir );
      strcat( FName, AreasFileName );
    }

    in.open( FName );

    if (in) {
      if (QuietMode == 0)
        cout << "Reading file " << FName << "...";
      while (!in.eof()) {
        in.getline( buf, MAX_LINE );
        if (strnicmp( buf, "Area ", 5) == 0) {
          for (i = 5; buf[i] == ' '; i++)
            ;
          for (j = i; (buf[j] != '\0') && (buf[j] != ' '); j++)
            ;
          if (j > i)
            j--;
          memcpy( Area, &buf[i], j-i+1 );
          Area[j-i+1] = '\0';
          
          for (i = strlen(buf)-1; (i >= 0) && (buf[i] == ' '); i--)
            ;
          if ((i >= 0) && (buf[i] == '"')) {
            for (j = i-1; (j >= 0) && (buf[j] != '"'); j--)
              ;
            if (j >= 0) {
              j++;
              i--;
              if (j <= i) {
                memcpy( Desc, &buf[j], i-j+1 );
                Desc[i-j+1] = '\0';
                // 饬 
                for (i = 0; i < AreasNumber; i++)
                  if (stricmp(Areas[i].Name, Area) == 0)
                    break;
                // 諨  騬 ᠭ?
                if ((i < AreasNumber) && (Areas[i].Desc == 0)) {
                  Areas[i].Desc = new char[strlen(Desc)+1];
                  strcpy( Areas[i].Desc, Desc );
                }
              }
            }
          }
        }
      }
      in.close();
      if (QuietMode == 0)
        cout << "Done.\n";
    }
    else {
      if (QuietMode == 0)
        cout << "[!] Failed reading file " << FName << endl
             << "[!] Area descriptions will be undefined.\n";
    }

    // 襬 

    if (ResultFileName[1] == ':')
      strcpy( FName, ResultFileName );
    else {
      strcpy( FName, AppDir );
      strcat( FName, ResultFileName );
    }

    out.open( FName );

    if (!out) {
      if (QuietMode == 0)
        cout << "[!] Failed writing report to " << FName << endl;
      for (i = 0; i < AreasNumber; i++) {
        for (j = 0; j < Areas[i].FilesNumber; j++) {
          delete [] Areas[i].Files[j].Name;
          delete [] Areas[i].Files[j].Desc;
          delete [] Areas[i].Files[j].Origin;
          delete [] Areas[i].Files[j].Uplink;
        }
        delete [] Areas[i].Files;
        delete [] Areas[i].Name;
        if (Areas[i].Desc != 0)
          delete [] Areas[i].Desc;
      }
      delete [] Areas;
  
      if (TplsNumber > 0) {
        for (i = 0; i < TplsNumber; i++)
          if (PathTpl[i].ID == ID_SUBSTRING)
            delete [] PathTpl[i].S;
        delete [] PathTpl;
      }

      return;
    }

    if (QuietMode == 0)
      cout << "Writing report to " << FName << "...";

    // if header was set, print it instead of information
    if (*HeaderFileName) {
        ifstream header;

        if (HeaderFileName[1] == ':')
            strcpy( FName, HeaderFileName );
        else {
            strcpy( FName, AppDir );
            strcat( FName, HeaderFileName );
        }
        header.open( FName );
        if (header) {
            while (!header.eof()) {
                header.getline( buf, MAX_LINE );
                if (header)
                    out << buf << endl;
            }
            header.close();
        }
    }
    else {
        // 뢮 ଠ  ᥡ
    
        time_t TimeOfDay;
        char   cTime[26];
    
        TimeOfDay = time( NULL );
    
        #ifdef __WATCOMC__
        _ctime( &TimeOfDay, cTime );
        #elif __BORLANDC__
        strcpy( cTime, ctime( &TimeOfDay ) );
        #endif
        cTime[24] = 0;
    
        // _F_ree f_I_lin fi_L_e rep_O_rter
        out << "Report generated on " << cTime << " by F.I.L.O. "
            << VerMajor << "." << VerMinor << " for Win32\n\n";
    }
    
    // 뢮 ⠫ 

    int  len, from, index, k;
    char SubString[81];

    TotalSize = 0;
    TotalNumber = 0;
    for (i = 0; i < AreasNumber; i++) {
      // ନ㥬   ᠭ , ⮡ ப  ⨫ 
      // 78 ᨬ, ..  "_<>_(ᠭ)_"  78 - 2 = 76 ᨬ-
      // ,   ஡ - 74 ᨬ
      //  Areas[i].Name   ॢ 74 ᨬ
      if (strlen(Areas[i].Name) > 74) {
        Areas[i].Name[74] = '\0';
        for (j = 73; j >= 71; j--)
          Areas[i].Name[j] = '.';
      }
      len = 74 - strlen(Areas[i].Name); // ᪮쪮 ᨬ ⠫
      // 뢮  
      out << "> " << Areas[i].Name;
      // ᠭ ?
      if (Areas[i].Desc != 0) {
        //    뢮 ᠭ?
        // 室   "_(<Desc>...)", .. 7 ᨬ
        // 3 㦥 + ᨬ ᠭ;
        //  2 ਠ
        if (((len - 3) >= strlen(Areas[i].Desc)) && (Areas[i].Desc[0] != '\0'))
        { // ᠭ ?
          // 뢮 ᠭ
          out << " (" << Areas[i].Desc << ")";
          len -= (strlen(Areas[i].Desc) + 3);
        }
        else { // ᠭ  
          if (len >= 7) { //    㪮稢?
            // ᠭ ᥪ  len-3 ᨬ
            Areas[i].Desc[len-3] = '\0';
            for (j = len-4; j >= len-6; j--)
              Areas[i].Desc[j] = '.';
            // 뢮
            out << " (" << Areas[i].Desc << ")";
            len -= (strlen(Areas[i].Desc) + 3);
          }
        }
      }
      out << " ";
      // 뢮 len+1 ᨬ ''
      for (j = len; j >= 0; j--)
        out << "";
      out << endl;
      AreaSize = 0;
      for (j = 0; j < Areas[i].FilesNumber; j++) {
        out << setiosflags(ios::left) << setw(14) << Areas[i].Files[j].Name;
        FormatSize( Areas[i].Files[j].Size, buf );
        out << resetiosflags(ios::left) << setiosflags(ios::right)
            << setw(10) << buf << " ";
        // 뢮 ᠭ
        len = strlen(Areas[i].Files[j].Desc);
        if (len <= 53) // 78-25
          out << setw(0) << Areas[i].Files[j].Desc << endl;
        else {
          index = GetSubstringLen( Areas[i].Files[j].Desc, 53 );
          strncpy( SubString, Areas[i].Files[j].Desc, index );
          SubString[index] = 0;
          out << setw(0) << SubString << endl;
          from = index;
          while (Areas[i].Files[j].Desc[from] != 0) {
            while (Areas[i].Files[j].Desc[from] == ' ')
              from++;
            if (Areas[i].Files[j].Desc[from] != 0) {
              out << SpaceLine;
              if ((len - from) <= 53) {
                out << &Areas[i].Files[j].Desc[from] << endl;
                from = len;
              }
              else {
                index = GetSubstringLen( &Areas[i].Files[j].Desc[from], 53 );
                strncpy( SubString, &Areas[i].Files[j].Desc[from], index );
                SubString[index] = 0;
                out << SubString << endl;
                from += index;
              }
            }
          }
        }
        if (TplsNumber > 0) {
          out << SpaceLine;
          for (k = 0; k < TplsNumber; k++) {
            switch (PathTpl[k].ID) {
              case ID_SUBSTRING:
                out << PathTpl[k].S;
                break;
              case ID_UPLINK:
                out << Areas[i].Files[j].Uplink;
                break;
              case ID_ORIGIN:
                out << Areas[i].Files[j].Origin;
                break;
            }
          }
          out << endl;
        }
        AreaSize += Areas[i].Files[j].Size;
      }
      out << ">";
      for (j = 1; j < 78; j++)
        out << "";
      out << endl;
      FormatSize( AreaSize, buf );
      out << "Total: " << Areas[i].FilesNumber << " file(s) of "
          << buf << " byte(s)\n";
      TotalNumber += Areas[i].FilesNumber;
      out << endl;
      TotalSize += AreaSize;
    }
    FormatSize( TotalSize, buf );

    out << "Total files number: " << TotalNumber << endl;
    out << "Total files size: " << buf << " byte(s)";

    if (AddEmptyLine)
      out << endl;

    if (*TailFileName) {
        ifstream tail;

        if (TailFileName[1] == ':')
            strcpy( FName, TailFileName );
        else {
            strcpy( FName, AppDir );
            strcat( FName, TailFileName );
        }
        tail.open( FName );
        if (tail) {
            while (!tail.eof()) {
                tail.getline( buf, MAX_LINE );
                if (tail)
                    out << buf << endl;
            }
            tail.close();
        }
    }
    
    out.close();

    if (QuietMode == 0)
      cout << "Done.\n";

    if (CreateFileFlags && AreasNumber > 0) {
        if (QuietMode == 0)
            cout << "Creating file flags...";
        
        int  len, CreatedNumber;

        // prepare flags directory name
        
        if (FileFlagsDir[1] == ':')
            strcpy( FName, FileFlagsDir );
        else {
            strcpy( FName, AppDir );
            strcat( FName, FileFlagsDir );
        }
        len = strlen(FName);
        if (FName[len-1] != '\\') {
            FName[len] = '\\';
            len++;
            FName[len] = '\0';
        }

        CreatedNumber = 0;
        for (i = 0; i < AreasNumber; i++) {
            // make flag name
            strcpy( &FName[len], Areas[i].Name );
            // create flag file with zero length
            if (CreateZeroFile( FName ))
                CreatedNumber++;
        }
        if (QuietMode == 0) {
            cout << "Done.\n";
            if (CreatedNumber < AreasNumber) {
                if (CreatedNumber == 0)
                    cout << "[!] None of flag files were created.\n";
                else
                    cout << "[!] Only " << CreatedNumber
                         << " flag files were created instead of "
                         << AreasNumber << endl;
                cout << "[!] Check FileFlagsDir parameter in "
                     << ConfigFileName << endl;
            }
        }
    }
    
    for (i = 0; i < AreasNumber; i++) {
      for (j = 0; j < Areas[i].FilesNumber; j++) {
        delete [] Areas[i].Files[j].Name;
        delete [] Areas[i].Files[j].Desc;
        delete [] Areas[i].Files[j].Origin;
        delete [] Areas[i].Files[j].Uplink;
      }
      delete [] Areas[i].Files;
      delete [] Areas[i].Name;
      if (Areas[i].Desc != 0)
        delete [] Areas[i].Desc;
    }
    delete [] Areas;

    if (TplsNumber > 0) {
      for (i = 0; i < TplsNumber; i++)
        if (PathTpl[i].ID == ID_SUBSTRING)
          delete [] PathTpl[i].S;
      delete [] PathTpl;
    }

  }

  void TruncateAddress(char *addr) {
    int i, j;

    for (i = 0; (addr[i] != 0) && (addr[i] != '/'); i++)
      if (addr[i] != Address[i])
        break;
    // ᮢ    ॣ
    if ((addr[i] == '/') && (Address[i] == '/')) {
      j = i+1;
      while (addr[j] != 0) {
        addr[j-i-1] = addr[j];
        j++;
      }
      addr[j-i-1] = 0;
    }
  }                     

  void SortFiles( FileStruct *Files, int FilesNumber ) {
    int   i, j, max;
    char *Name;
    int   Size;
    char *Desc;
    char *Origin;
    char *Uplink;

    for (i = FilesNumber-1; i > 0; i--) {
      max = 0;
      for (j = 1; j <= i; j++)
        if (stricmp( Files[j].Name, Files[max].Name ) > 0)
          max = j;
      if (max != i) {
        Name = Files[i].Name;
        Size = Files[i].Size;
        Desc = Files[i].Desc;
        Origin = Files[i].Origin;
        Uplink = Files[i].Uplink;
        Files[i].Name = Files[max].Name;
        Files[i].Size = Files[max].Size;
        Files[i].Desc = Files[max].Desc;
        Files[i].Origin = Files[max].Origin;
        Files[i].Uplink = Files[max].Uplink;
        Files[max].Name = Name;
        Files[max].Size = Size;
        Files[max].Desc = Desc;
        Files[max].Origin = Origin;
        Files[max].Uplink = Uplink;
      }
    }
  }

  void SortAreas( AreaStruct *Areas, int AreasNumber ) {
    int   i, j, max;
    char       *Name;
    char       *Desc;
    FileStruct *Files;
    int         FilesNumber;

    for (i = AreasNumber-1; i > 0; i--) {
      max = 0;
      for (j = 1; j <= i; j++)
        if (strcmp( Areas[j].Name, Areas[max].Name) > 0)
          max = j;
      if (max != i) {
        Name = Areas[i].Name;
        Desc = Areas[i].Desc;
        Files = Areas[i].Files;
        FilesNumber = Areas[i].FilesNumber;
        Areas[i].Name = Areas[max].Name;
        Areas[i].Desc = Areas[max].Desc;
        Areas[i].Files = Areas[max].Files;
        Areas[i].FilesNumber = Areas[max].FilesNumber;
        Areas[max].Name = Name;
        Areas[max].Desc = Desc;
        Areas[max].Files = Files;
        Areas[max].FilesNumber = FilesNumber;
      }
    }
  }

  // 頥  ப
  int GetSubstringLen(char *S, int width) {
    int i;

    for (i = width-1; (i >= 0) && (S[i] != ' '); i--)
      ;
    if (i < 0)
      return (width);
    return (i+1);
  }

  // - ଠ஢ ࠧ
  void FormatSize(int size, char *buf) {
    int i, j, len, count;

    sprintf( buf, "%ld", size );
    len = strlen(buf);

    count = 0; 
    for (i = len-1; i >= 0; i--) {
      count++;
      if ((count == 3) && (i > 0)) { // ⠢塞 ','
        count = 0;
        // ᤢ ⮪ ࠢ
        for (j = len+1; j > i; j--)
          buf[j] = buf[j-1];
        len++;
        buf[i] = ',';
      }
    }
  }

  void PrepareTemplates() {
    struct TplListStruct {
      TemplateStruct  Tpl;
      TplListStruct   *Next;
    };
    TplListStruct *TplList, *LastTpl;
    int            i, j;

    TplsNumber = 0;
    TplList = 0;
    i       = 0;
    j       = 0;
    while (PathTemplate[j] != 0) {
      if (PathTemplate[j] == '@') {
        j++;
        switch (PathTemplate[j]) {
          case 'U':
          case 'u':
            //   ப?
            if ((j - i) > 1) {
              if (TplList == 0) {
                TplList = new TplListStruct;
                LastTpl = TplList;
              }
              else {
                LastTpl->Next = new TplListStruct;
                LastTpl = LastTpl->Next;
              }
              LastTpl->Next = 0;
              LastTpl->Tpl.ID = ID_SUBSTRING;
              LastTpl->Tpl.S = new char[j - i];
              memcpy( LastTpl->Tpl.S, &PathTemplate[i], j-i-1 );
              LastTpl->Tpl.S[j-i-1] = '\0';
              TplsNumber++;
            }
            // 塞 Uplink-
            if (TplList == 0) {
              TplList = new TplListStruct;
              LastTpl = TplList;
            }
            else {
              LastTpl->Next = new TplListStruct;
              LastTpl = LastTpl->Next;
            }
            LastTpl->Next = 0;
            LastTpl->Tpl.ID = ID_UPLINK;
            TplsNumber++;
            j++;
            i = j;
            break;
          case 'O':
          case 'o':
            //   ப?
            if ((j - i) > 1) {
              if (TplList == 0) {
                TplList = new TplListStruct;
                LastTpl = TplList;
              }
              else {
                LastTpl->Next = new TplListStruct;
                LastTpl = LastTpl->Next;
              }
              LastTpl->Next = 0;
              LastTpl->Tpl.ID = ID_SUBSTRING;
              LastTpl->Tpl.S = new char[j - i];
              memcpy( LastTpl->Tpl.S, &PathTemplate[i], j-i-1 );
              LastTpl->Tpl.S[j-i-1] = '\0';
              TplsNumber++;
            }
            // 塞 Origin
            if (TplList == 0) {
              TplList = new TplListStruct;
              LastTpl = TplList;
            }
            else {
              LastTpl->Next = new TplListStruct;
              LastTpl = LastTpl->Next;
            }
            LastTpl->Next = 0;
            LastTpl->Tpl.ID = ID_ORIGIN;
            TplsNumber++;
            j++;
            i = j;
            break;
          default:
            j++;
        }
      }
      else
        j++;
    }
    //   ப?
    if ((j - i) > 0) {
      if (TplList == 0) {
        TplList = new TplListStruct;
        LastTpl = TplList;
      }
      else {
        LastTpl->Next = new TplListStruct;
        LastTpl = LastTpl->Next;
      }
      LastTpl->Next = 0;
      LastTpl->Tpl.ID = ID_SUBSTRING;
      LastTpl->Tpl.S = new char[j - i + 1];
      memcpy( LastTpl->Tpl.S, &PathTemplate[i], j-i );
      LastTpl->Tpl.S[j-i] = '\0';
      TplsNumber++;
    }
    
    if (TplsNumber > 0) {
      PathTpl = new TemplateStruct[TplsNumber];
      for (i = 0; i < TplsNumber; i++) {
        PathTpl[i].S = TplList->Tpl.S;
        PathTpl[i].ID = TplList->Tpl.ID;
        LastTpl = TplList->Next;
        delete TplList;
        TplList = LastTpl;
      }
    }
  }                        

  char SkipTable[256];

  void PrepareSkipTable() {
    int  i, j;
    int  LastChar; // ᫥ ⠭ ᨬ, -1, ᫨  ⠭
    int  WaitRangeEnd; // 1 -  ᨬ  , -1 - 㦥 
    int  RangeEnd; // ᨬ  
    int  Error;

    // 塞 ⠡ (..  ᨬ ᯮ)
    memset( SkipTable, 0, sizeof(SkipTable));

    // 㥬 ப SkipSet
    i = 0;
    LastChar = -1;
    WaitRangeEnd = 0;
    Error = 0;
    while ((SkipSet[i] != 0) && (Error == 0)) {
      switch (SkipSet[i]) {
        case ' ':
          i++;
          break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          if (WaitRangeEnd == 1) {
            RangeEnd = SkipSet[i] - '0';
            i++;
            while ((SkipSet[i] >= '0') && (SkipSet[i] <= '9')) {
              RangeEnd *= 10;
              RangeEnd += (SkipSet[i] - '0');
              i++;
            } 
            if (RangeEnd > 255) {
              Error = 1;
              break;
            }
            if (RangeEnd < LastChar) {
              Error = 1; // 訡
              break;
            }
            for (j = LastChar; j < RangeEnd; j++)
              SkipTable[j] = 1;
            WaitRangeEnd = -1;
          }
          else {
            if (LastChar != -1) {
              Error = 1; // 訡
              break;
            }
            LastChar = SkipSet[i] - '0';
            i++;
            while ((SkipSet[i] >= '0') && (SkipSet[i] <= '9')) {
              LastChar *= 10;
              LastChar += (SkipSet[i] - '0');
              i++;
            }
            if (LastChar > 255) {
              Error = 1;
              break;
            }
          }
          break;
        case ',':
          if ((LastChar == -1) || (WaitRangeEnd == 1)) {
            Error = 1;
            break;
          }
          SkipTable[LastChar] = 1;
          LastChar = -1;
          WaitRangeEnd = 0;
          i++;
          break;
        case '-':
          if ((LastChar == -1) || (WaitRangeEnd != 0)) {
            Error = 1;
            break;
          }
          WaitRangeEnd = 1;
          i++;
          break;
        case 39: // '
          i++;
          if ((SkipSet[i] == '\0') || (SkipSet[i+1] != 39)) {
            Error = 1; // 訡
            break;
          }
          if (WaitRangeEnd == 1) {
            RangeEnd = SkipSet[i];
            if (RangeEnd < LastChar) {
              Error = 1; // 訡
              break;
            }
            for (j = LastChar; j < RangeEnd; j++)
              SkipTable[j] = 1;
            WaitRangeEnd = -1;
          }
          else {
            if (LastChar != -1) {
              Error = 1; // 訡
              break;
            }
            LastChar = SkipSet[i];
          }
          i++;
          i++;
          break;
        default:
          Error = 1; // 訡
          break;
      }
    }
    if (Error || (LastChar == -1) || (WaitRangeEnd == 1)) {
      if (Error)
          cout << "[!] Error parsing skip set. Chars will not be skipped.\n"
               << "[!] Check SkipSet parameter in " << ConfigFileName << endl;
      SkipDescChars = 0;
    }
    else
      SkipTable[LastChar] = 1;
  }

  void DoSkipChars(char *S) {
    int   i, j;
    char *buf;

    buf = new char[strlen(S) + 1];
    i = 0;
    j = 0;
    while (S[i] != '\0') {
      if (SkipTable[S[i]] == 0) {
        buf[j] = S[i];
        j++;
      }
      i++;
    }
    buf[j] = '\0';
    strcpy( S, buf );

    delete [] buf;
  }

  void DoReplaceSoftCR(char *S) {
    int i;

    i = 0;
    while (S[i] != 0) {
      if (S[i] == 141) // ᪠  ()
        S[i] = 72;     // ᪠ H ()
      i++;
    }
  }

  void SkipSpaces(char *s) {
      char *head;
      char *tail;

      head = s;
      while (*head == ' ')
          head++;
      if (*head) {
          tail = head + strlen(head) - 1;
          while (*tail == ' ')
              tail--;
          tail++;
          *tail = '\0';
      }
      while (*head) {
          *s = *head;
          s++;
          head++;
      }
      *s = '\0';
  }
  
  void PrintLogo() {
    cout << "F.I.L.O. " << VerMajor << "." << VerMinor << " for Win32\n";
    if (QuietMode == 0) {
      cout << "Usage: freport.exe [-q]\n"
           << "Use -q option for quiet mode (only logo will be printed).\n";
    }
  }

  int CreateZeroFile( char *fname ) {
      FILE *file;

      file = fopen( fname, "wb");
      if (file) {
          fclose( file );
          return 1;
      }
      // failed
      return 0;
  }